unit list_utils;
{
 DESCRIPTION     :  Unit providing UI-neutral routines related to listing,
                    counting and check accessibility of objects in a given path

 REQUIREMENTS    :  FPC

 EXTERNAL DATA   :  ---

 MEMORY USAGE    :  ---

 DISPLAY MODE    :  ---

 REFERENCES      :  ---

 REMARK          :  ---

 Version  Date      Author      Modification
 -------  --------  -------     ------------------------------------------
 0.10     20060831  G.Tani      Initial version
 0.11     20060917  G.Tani      removed *_VER; P_RELEASE constant in pea_utils
                                is used to keep track of release level;
                                changed extractdirname: if the path is a root (i.e.
                                disk name in Windows) the path becomes the dir path
                                and the dir name is blanked;
 0.12     20060925  G.Tani      fixed a bug in extractdirname showing up only
                                with Lazarus 0.9.18
 0.13     20080721  G.Tani      Enabled utf8
 0.14     20080825  G.Tani      Added srcfilesize to get file size using TSearchRec
 0.15     20090430  G.Tani      Added rCount for counting files in a path (optionally recursive)
 0.16     20090530  G.Tani      Added srcfilesize_multipart to get size of spanned archives (found on the same path of .001 file)
 0.17     20090601  G.Tani      Added rCountSize for counting number of files and total size in a path (optionally recursive)

(C) Copyright 2006 Giorgio Tani giorgiotani@interfree.it
The program is released under GNU LGPL http://www.gnu.org/licenses/lgpl.txt

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 3 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
}

{$mode objfpc}{$H+}
{$INLINE ON}

interface

uses Classes, SysUtils, StrUtils, ansiutf8_utils;

type
   TFoundList = array of utf8string; //dynamic array of utf8string, each time a new file is found new space is allocated (function rList)
   TFoundListBool = array of boolean;
   TFoundListAges = array of dword;
   TFoundListAttrib = array of dword;
   TFoundListSizes = array of qword;
   TFoundListArray64 = array of array [0..63] of byte;
   TFileOfByte = File of Byte;

const
SUCCESS = 0;
INCOMPLETE_FUNCTION = 1;
LIST_ERROR = 2;
LIST_RECURSION_ERROR = 3;
CALL_ERROR = 4;
DWORD_DECODE_TO_ATTRIBUTES_ERROR = 5;
STRING_DECODE_TO_ATTRIBUTES_ERROR = 6;
OBJECT_NOT_ACCESSIBLE = 7;

//optionally recursive fucntion to count files
function rCount  ( dir,mask: utf8string;                                        //path and mask for search
                  fattrib: qword;                                               //file attributes to filter the search
                  recur: boolean;                                               //true: recursive (search in all subfolders); false not recursive (search only in main folder)
                  var nfiles,ndirs: qword                                       //number of files and dirs found
                  ):integer;

//optionally recursive fucntion to count files and total size
function rCountSize  ( dir,mask: utf8string;                                    //path and mask for search
                     fattrib: qword;                                            //file attributes to filter the search
                     recur: boolean;                                            //true: recursive (search in all subfolders); false not recursive (search only in main folder)
                     var nfiles,ndirs,tsize: qword                              //number of files and dirs found, total size
                     ):integer;

{
if an input object is accessible, assign it as a file of byte
otherwise mark it as skipped at the given address in the status_files TFoundList
}
function check_in ( var f_in:TFileOfByte;                                       //file of byte to assign
                    in_qualified_name: utf8string;                              //qualified name of the object to test
                    var status_files:TFoundListBool;                            //TFoundListBool containing accessibility boolean value of the object to test
                    addr:dword                                                  //address of the object to test in the TFoundList
                    ):integer;

function srcfilesize ( s:utf8string;
                       var fsize:qword):integer;

function srcfilesize_multipart ( s:utf8string;
                                 var fsize:qword):integer;

{
list files, dirs and size matching the given path, mask and file attributes;
optionally include main dir as first object;
optionally list object with or without recursion
}
function dirsize ( path,mask: utf8string;                                       //path and mask for search
                   fattrib: qword;                                              //file attributes to filter the search
                   recur: boolean;                                              //if true: recursive listing (search in all subfolders); if false: not recursive listing (search in main folder)
                   addtofirstobj: boolean;                                      //if true start to update list from second element (the first one is left available for list the main object, i.e. list folder name and then the content of the folder); if false start listing from first array element (0, list content of the folder)
                   var nfiles,ndirs: qword;                                     //number of files and dirs found
                   var size:qword                                               //total size (Byte)
                   ):integer;

{
decode file attributes given attributes as a dword
}
function dword2decodedFileAttributes ( d: dword;                                //input dword
                                       var fattrib:utf8string                   //string of file attributes abbreviations
                                       ): integer;

{
extract directory name and path were the directory is in (nesting level
unlimited by implementation, only memory limited, whidely exceeding needs)
}
function extractdirname ( inpath:utf8string;                                    //input path
                          var dirpath:utf8string;                               //path were the directory is in
                          var dirname:utf8string                                //name of the dir
                          ):integer;

function ansiextractdirname ( inpath:ansistring;                                    //input path
                              var dirpath:ansistring;                               //path were the directory is in
                              var dirname:ansistring                                //name of the dir
                              ):integer;
                              
{
expand an object, given it's name, to a list of objects (TFoundList).
If the object is a file, its name is returned as sole element of the list;
otherwise if the object is a directory it's name and, recursively, all the
content, will be added to the list
}
function expand ( s:utf8string;                                                 //input object
                  var exp_files:TFoundList;                                     //expanded list of objects
                  var exp_fsizes:TFoundListSizes;                               //expanded list of object sizes
                  var exp_ftimes:TFoundListAges;                                //expanded list of object ages
                  var exp_fattr:TFoundListAttrib;                               //expanded list of object attributes
                  var exp_fattr_dec:TFoundList;                                 //expanded list of object decoded attributes
                  var nfound:qword                                              //number of objects found
                  ):integer;                                                    //take the exit code from ListFile function that it calls

{
list files and optionally dirs in a path (optionally with recursion in subdirs),
list objects details to other separate lists
}
function listdetails ( path,mask: utf8string;                                   //path and mask for search
                       fattrib: qword;                                          //file attributes to filter the search
                       recur: boolean;                                          //if true: recursive listing (search in all subfolders); if false: not recursive listing (search in main folder)
                       addtofirstobj: boolean;                                  //if true start to update list from second element (the first one is left available for list the main object, i.e. list folder name and then the content of the folder); if false start listing from first array element (0, list content of the folder)
                       var nfiles,ndirs: qword;                                 //number of files and dirs found
                       var flist: TFoundList;                                   //list of file names
                       var fsizes: TFoundListSizes;                             //list of file sizes
                       var ftimes: TFoundListAges;                              //list of file ages
                       var fattr: TFoundListAttrib;                             //list of file attributes
                       var fattr_dec: TFoundList                                //list of file decoded attributes
                       ):integer;

implementation

//optionally recursive function used internally to list files (optionally including dirs) and object details
function rList  ( mode:string;                                                  //mode of operation, currently implemented only DETAILS: give detailed output (1+4 lists);
                  dir,mask: utf8string;                                         //path and mask for search
                  fattrib: qword;                                               //file attributes to filter the search
                  recur: boolean;                                               //true: recursive (search in all subfolders); false not recursive (search only in main folder)
                  var nfiles,ndirs: qword;                                      //number of files and dirs found
                  var flist: TFoundList;                                        //list of file names
                  var fsizes: TFoundListSizes;                                  //list of file sizes
                  var ftimes: TFoundListAges;                                   //list of file ages
                  var fattr: TFoundListAttrib;                                  //list of file attributes
                  var fattr_dec: TFoundList                                     //list of file decoded attributes
                  ):integer;
var r: TSearchRec;
begin
rList:=INCOMPLETE_FUNCTION;
inc(ndirs,1);
if uFindFirst(dir + mask, fattrib, r) = 0 then begin
   try
   if upcase(mode)='DETAILS' then
      repeat
         if ((r.Name <> '.') and (r.Name <> '..')) then
            begin
            SetLength(flist,nfiles+1);
            SetLength(fsizes,nfiles+1);
            SetLength(ftimes,nfiles+1);
            SetLength(fattr,nfiles+1);
            SetLength(fattr_dec,nfiles+1);
            flist[nfiles]:=dir + envtoutf8(r.Name);
            if (r.Attr and faDirectory <> 0) then
               if flist[nfiles][length(flist[nfiles])]<>DirectorySeparator then flist[nfiles]:=flist[nfiles]+DirectorySeparator;
            fsizes[nfiles]:=r.Size;
            ftimes[nfiles]:=r.Time;
            fattr[nfiles]:=r.Attr;
            dword2decodedFileAttributes(r.Attr,fattr_dec[nfiles]);
            inc(nfiles,1);
            end;
      until FindNext(r) <> 0;
   except
      FindClose(r);
      rList:=LIST_ERROR;
      exit;
   end;
   FindClose(r);
end;
if recur=true then
   if uFindFirst(dir + '*', fattrib Or faDirectory, r) = 0 then
      begin
      try
         repeat
            if ((r.Attr and faDirectory) <> 0) and (r.Name <> '.') and (r.Name <> '..') then rList(mode,dir + envtoutf8(r.Name) + DirectorySeparator,mask,fattrib,recur,nfiles,ndirs,flist,fsizes,ftimes,fattr,fattr_dec);
         until FindNext(r) <> 0;
      except
         FindClose(r);
         rList:=LIST_RECURSION_ERROR;
         exit;
      end;
      FindClose(r);
      end;
if rList=INCOMPLETE_FUNCTION then rList:=SUCCESS;
end;

//optionally recursive fucntion to count files
function rCount  ( dir,mask: utf8string;                                         //path and mask for search
                  fattrib: qword;                                               //file attributes to filter the search
                  recur: boolean;                                               //true: recursive (search in all subfolders); false not recursive (search only in main folder)
                  var nfiles,ndirs: qword                                      //number of files and dirs found
                  ):integer;
var r: TSearchRec;
begin
rCount:=INCOMPLETE_FUNCTION;
inc(ndirs,1);
if uFindFirst(dir + mask, fattrib, r) = 0 then begin
   try
      repeat
         if ((r.Name <> '.') and (r.Name <> '..')) then
            begin
            inc(nfiles,1);
            end;
      until FindNext(r) <> 0;
   except
      FindClose(r);
      rCount:=LIST_ERROR;
      exit;
   end;
   FindClose(r);
end;
if recur=true then
   if uFindFirst(dir + '*', fattrib Or faDirectory, r) = 0 then
      begin
      try
         repeat
            if ((r.Attr and faDirectory) <> 0) and (r.Name <> '.') and (r.Name <> '..') then rCount(dir + envtoutf8(r.Name) + DirectorySeparator,mask,fattrib,recur,nfiles,ndirs);
         until FindNext(r) <> 0;
      except
         FindClose(r);
         rCount:=LIST_RECURSION_ERROR;
         exit;
      end;
      FindClose(r);
      end;
if rCount=INCOMPLETE_FUNCTION then rCount:=SUCCESS;
end;

//optionally recursive fucntion to count files and total size
function rCountSize  ( dir,mask: utf8string;                                    //path and mask for search
                  fattrib: qword;                                               //file attributes to filter the search
                  recur: boolean;                                               //true: recursive (search in all subfolders); false not recursive (search only in main folder)
                  var nfiles,ndirs,tsize: qword                                 //number of files and dirs found, total size
                  ):integer;
var r: TSearchRec;
begin
rCountSize:=INCOMPLETE_FUNCTION;
inc(ndirs,1);
if uFindFirst(dir + mask, fattrib, r) = 0 then begin
   try
      repeat
         if ((r.Name <> '.') and (r.Name <> '..')) then
            begin
            inc(nfiles,1);
            tsize:=tsize+r.Size;
            end;
      until FindNext(r) <> 0;
   except
      FindClose(r);
      rCountSize:=LIST_ERROR;
      exit;
   end;
   FindClose(r);
end;
if recur=true then
   if uFindFirst(dir + '*', fattrib Or faDirectory, r) = 0 then
      begin
      try
         repeat
            if ((r.Attr and faDirectory) <> 0) and (r.Name <> '.') and (r.Name <> '..') then rCountSize(dir + envtoutf8(r.Name) + DirectorySeparator,mask,fattrib,recur,nfiles,ndirs,tsize);
         until FindNext(r) <> 0;
      except
         FindClose(r);
         rCountSize:=LIST_RECURSION_ERROR;
         exit;
      end;
      FindClose(r);
      end;
if rCountSize=INCOMPLETE_FUNCTION then rCountSize:=SUCCESS;
end;

function check_in ( var f_in:TFileOfByte;                                       //file of byte to assign
                    in_qualified_name: utf8string;                              //qualified name of the object to test
                    var status_files:TFoundListBool;                            //TFoundListBool containing accessibility boolean value of the object to test
                    addr:dword                                                  //address of the object to test in the TFoundList
                    ):integer;
var
   in_ok:boolean;
begin
check_in:=INCOMPLETE_FUNCTION;
status_files[addr]:=false;
try
   if ufilegetattr(in_qualified_name) > 0 then
      if ufilegetattr(in_qualified_name) and faDirectory =0 then
         begin
         uassignfile(f_in,in_qualified_name);
         filemode:=0;
         reset(f_in);
         in_ok:=true;
         end
      else in_ok:=true
   else in_ok:=false;
except
   in_ok:=false;
end;
if in_ok=false then
   begin
   status_files[addr]:=false;
   check_in:=OBJECT_NOT_ACCESSIBLE;
   end
else
   begin
   status_files[addr]:=true;
   check_in:=SUCCESS;
   end;
end;

function srcfilesize ( s:utf8string;
                       var fsize:qword):integer;
var
   r: TSearchRec;
begin
srcfilesize:=-1;
if ufilegetattr(s) and faDirectory = 0 then //object is a file
   begin
   if uFindFirst(s,faAnyFile,r) = 0 then
      begin
      fsize:=r.Size;
      srcfilesize:=0;
      end;
   FindClose(r);
   end;
end;

function srcfilesize_multipart ( s:utf8string;
                                 var fsize:qword):integer;
var
   s_ext,s_name,s_path:utf8string;
   i,j:integer;
   r: TSearchRec;
begin
srcfilesize_multipart:=-1;
fsize:=0;
if ufilegetattr(s) and faDirectory = 0 then //object is a file
   begin
   s_ext:=uextractfileext(s);
   s_name:=uextractfilename(s);
   setlength(s_name,length(s_name)-length(s_ext));
   s_path:=uextractfilepath(s);
   i:=1;
   if s_ext='.001' then
      repeat
         j:=uFindFirst(s_path+s_name+s_ext,faAnyFile,r);
         if j=0 then fsize:=fsize+r.Size;
         FindClose(r);
         i:=i+1;
         if i<10 then s_ext:='.00'+inttostr(i)
         else
            if i<100 then s_ext:='.0'+inttostr(i)
            else s_ext:='.'+inttostr(i);
      until j<>0
   else
      begin
      if uFindFirst(s,faAnyFile,r) = 0 then
         begin
         fsize:=r.Size;
         srcfilesize_multipart:=0;
         end;
      FindClose(r);
      end;
   end;
end;

function dirsize ( path,mask: utf8string;                                       //path and mask for search
                   fattrib: qword;                                              //file attributes to filter the search
                   recur: boolean;                                              //if true: recursive listing (search in all subfolders); if false: not recursive listing (search in main folder)
                   addtofirstobj: boolean;                                      //if true start to update list from second element (the first one is left available for list the main object, i.e. list folder name and then the content of the folder); if false start listing from first array element (0, list content of the folder)
                   var nfiles,ndirs: qword;                                     //number of files and dirs found
                   var size:qword                                               //total size (Byte)
                   ):integer;
var
   i:integer;
   dir: utf8string;
   flist: TFoundList;
   fsizes: TFoundListSizes;
   ftimes: TFoundListAges;
   fattr: TFoundListAttrib;
   fattr_dec: TFoundList;
begin
DirSize:=INCOMPLETE_FUNCTION;
if addtofirstobj=true then nfiles:=1
else nfiles:=0;
ndirs:=0;
size:=0;
try
   //check for valid path, else set executable path as path
   if length(path)=0 then dir:=uextractfilepath(paramstr(0)) else dir:=path;
   //check for directoryseparator at the end of the path (needed)
   if dir[length(dir)]<>DirectorySeparator then dir:=dir+DirectorySeparator;
   DirSize:=rList('DETAILS',dir,mask,fattrib,recur,nfiles,ndirs,flist,fsizes,ftimes,fattr,fattr_dec);
   if addtofirstobj=false then ndirs:=ndirs-1;
   for i:=0 to length(fsizes)-1 do size:=size+fsizes[i];
except
   DirSize:=CALL_ERROR;
   exit;
end;
if DirSize=INCOMPLETE_FUNCTION then DirSize:=SUCCESS;
end;

function dword2decodedFileAttributes ( d: dword;                                //input dword
                                       var fattrib:utf8string                   //string of file attributes abbreviations
                                       ): integer;
begin
dword2decodedFileAttributes:=INCOMPLETE_FUNCTION;
try
   fattrib:='';
   if (d and $01) <> 0 then fattrib:=fattrib+'R';
   if (d and $02) <> 0 then fattrib:=fattrib+'H';
   if (d and $04) <> 0 then fattrib:=fattrib+'S';
   if (d and $08) <> 0 then fattrib:=fattrib+'V';
   if (d and $10) <> 0 then fattrib:=fattrib+'D';
   if (d and $20) <> 0 then fattrib:=fattrib+'A';
except
   dword2decodedFileAttributes:=DWORD_DECODE_TO_ATTRIBUTES_ERROR;
end;
if dword2decodedFileAttributes=INCOMPLETE_FUNCTION then dword2decodedFileAttributes:=SUCCESS;
end;

function extractdirname ( inpath:utf8string;                                    //input path
                          var dirpath:utf8string;                               //path were the directory is in
                          var dirname:utf8string                                //name of the dir
                          ):integer;
var s:utf8string;
i,j:integer;
dirarr:array of utf8string;
begin
extractdirname:=INCOMPLETE_FUNCTION;
s:=uextractfilepath(inpath);
i:=0;
while length(s)>0 do
   begin
   setlength(dirarr,i+1);
   dirarr[i]:=copy2symbdel(s,directoryseparator);
   if length(s)>0 then if s[1]=directoryseparator then s:=copy(s,2,length(s)-1);
   inc(i,1);
   end;
dirname:=dirarr[i-1];
dirpath:='';
for j:=0 to i-2 do dirpath:=dirpath+dirarr[j]+directoryseparator;
if dirpath='' then begin dirpath:=dirname; dirname:=''; end;
extractdirname:=SUCCESS;
end;

function ansiextractdirname ( inpath:ansistring;                                    //input path
                              var dirpath:ansistring;                               //path were the directory is in
                              var dirname:ansistring                                //name of the dir
                              ):integer;
var s:ansistring;
i,j:integer;
dirarr:array of ansistring;
begin
ansiextractdirname:=INCOMPLETE_FUNCTION;
s:=extractfilepath(inpath);
i:=0;
while length(s)>0 do
   begin
   setlength(dirarr,i+1);
   dirarr[i]:=copy2symbdel(s,directoryseparator);
   if length(s)>0 then if s[1]=directoryseparator then s:=copy(s,2,length(s)-1);
   inc(i,1);
   end;
dirname:=dirarr[i-1];
dirpath:='';
for j:=0 to i-2 do dirpath:=dirpath+dirarr[j]+directoryseparator;
if dirpath='' then begin dirpath:=dirname; dirname:=''; end;
ansiextractdirname:=SUCCESS;
end;

function expand ( s:utf8string;                                                 //input object
                  var exp_files:TFoundList;                                     //expanded list of objects
                  var exp_fsizes:TFoundListSizes;                               //expanded list of object sizes
                  var exp_ftimes:TFoundListAges;                                //expanded list of object ages
                  var exp_fattr:TFoundListAttrib;                               //expanded list of object attributes
                  var exp_fattr_dec:TFoundList;                                 //expanded list of object decoded attributes
                  var nfound:qword                                              //number of objects found
                  ):integer;                                                    //take the exit code from ListFile function that it calls
var
   i,j:qword;
   r:TSearchRec;
begin
nfound:=0;
i:=0;
j:=0;
SetLength(exp_files,i+1);
SetLength(exp_fsizes,i+1);
SetLength(exp_ftimes,i+1);
SetLength(exp_fattr,i+1);
SetLength(exp_fattr_dec,i+1);
exp_files[0]:=s;
dword2decodedFileAttributes(ufilegetattr(s),exp_fattr_dec[0]);
if ufilegetattr(s) and faDirectory = 0 then //object is a file
   begin
   if uFindFirst(s,faAnyFile,r) = 0 then
      begin
      exp_fsizes[0]:=r.Size;
      exp_ftimes[0]:=r.Time;
      exp_fattr[0]:=r.Attr;
      dword2decodedFileAttributes(r.Attr,exp_fattr_dec[0]);
      end;
   FindClose(r);
   end
else
   begin
   if exp_files[0][length(exp_files[0])]<>DirectorySeparator then exp_files[0]:=exp_files[0]+DirectorySeparator;
   exp_fsizes[0]:=0;
   if uFindFirst(s+'.',faDirectory,r) = 0 then
      begin
      exp_fsizes[0]:=r.Size;
      exp_ftimes[0]:=r.Time;
      exp_fattr[0]:=r.Attr;
      dword2decodedFileAttributes(r.Attr,exp_fattr_dec[0]);
      end;
   FindClose(r);
   expand := ListDetails ( s,'*',
                           faAnyFile,
                           true, //recursive listing
                           true, //add objects in list after the main one
                           i,j,
                           exp_files,exp_fsizes,exp_ftimes,exp_fattr,exp_fattr_dec);
   end;
nfound:=i;
end;

function listdetails ( path,mask: utf8string;                                   //path and mask for search
                       fattrib: qword;                                          //file attributes to filter the search
                       recur: boolean;                                          //if true: recursive listing (search in all subfolders); if false: not recursive listing (search in main folder)
                       addtofirstobj: boolean;                                  //if true start to update list from second element (the first one is left available for list the main object, i.e. list folder name and then the content of the folder); if false start listing from first array element (0, list content of the folder)
                       var nfiles,ndirs: qword;                                 //number of files and dirs found
                       var flist: TFoundList;                                   //list of file names
                       var fsizes: TFoundListSizes;                             //list of file sizes
                       var ftimes: TFoundListAges;                              //list of file ages
                       var fattr: TFoundListAttrib;                             //list of file attributes
                       var fattr_dec: TFoundList                                //list of file decoded attributes
                       ):integer;
var dir: utf8string;
begin
ListDetails:=INCOMPLETE_FUNCTION;
if addtofirstobj=true then nfiles:=1
else nfiles:=0;
ndirs:=0;
try
   //check for valid path, else set executable path as path
   if length(path)=0 then dir:=uextractfilepath(paramstr(0)) else dir:=path;
   //check for directoryseparator at the end of the path (needed)
   if dir[length(dir)]<>DirectorySeparator then dir:=dir+DirectorySeparator;
   ListDetails:=rList('DETAILS',dir,mask,fattrib,recur,nfiles,ndirs,flist,fsizes,ftimes,fattr,fattr_dec);
   if addtofirstobj=false then ndirs:=ndirs-1;
except
   ListDetails:=CALL_ERROR;
   exit;
end;
if ListDetails=INCOMPLETE_FUNCTION then ListDetails:=SUCCESS;
end;

end.

